use actix_web::web::Redirect;
use actix_web::{get, web, Responder};
use serde::Deserialize;


use sqlx::MySqlPool;
use std::{collections::HashMap, env};

use crate::joint_err::JointErr;




use crate::sql_users::{
    add_user_dont_care, get_refresh_token_by_id, AddUserSuccess,
};

const URL: &str = "https://accounts.spotify.com/api/token";

fn read_from_env() -> (String, String, String) {
    let id = env::var("CLIENT_ID").expect("Cannot proceed without client id");
    let secret = env::var("CLIENT_SECRET").expect("Cannot proceed without client secret");
    let url = env::var("REDIRECT_URL").expect("Cannot proceed without redirect url");
    (id, secret, url)
}

#[derive(Deserialize)]
pub struct CallbackData {
    code: String,
}

#[get("/callback")]
pub async fn callback_manager(
    data: web::Query<CallbackData>,
    state: web::Data<crate::state::AppState>,
) -> impl Responder {
    log::warn!("Access to /callback");
    let pool = &state.pool;

    let tokens = get_auth_token(&data.code).await;
    if let Ok(tokens) = tokens {
        log::info!("{:?}", tokens.refresh_token);
        let user_profile = get_user_email(&tokens.access_token).await;
        if let Ok(up) = user_profile {
            let q = add_user_dont_care(&up.email, &up.id, &tokens.refresh_token, pool).await;
            match q {
                Ok(AddUserSuccess::New) => {
                    log::warn!("Added a new user: {}", up.id);
                    Redirect::to(format!("/refresh?spotify_id={}&force=true", up.id))
                }
                Ok(AddUserSuccess::Duplicate) => {
                    Redirect::to(format!("/refresh?spotify_id={}", up.id))
                }
                _ => {
                    log::error!("SQL Query Failed: couldn't add new user\n{:?}", q);
                    Redirect::to("/failed")
                }
            }
        } else {
            log::error!(
                "We failed to get a user profile, the error was:\n{:?}",
                user_profile.unwrap_err()
            );
            Redirect::to("/failed")
        }
    } else {
        log::error!(
            "We failed to get tokens, the error was:\n{:?}",
            tokens.unwrap_err()
        );
        Redirect::to("/failed")
    }
}

#[derive(Deserialize, Debug)]
#[non_exhaustive]
struct AuthorizationRequestResponse {
    access_token: Token,
    refresh_token: Token,
}

#[derive(Deserialize, Debug)]
pub struct Token(pub String);

async fn get_auth_token(code: &str) -> Result<AuthorizationRequestResponse, reqwest::Error> {
    // Consider using env for REDIRECT_URI
    //const REDIRECT_URI: &'static str = "https://ilia.moe/cool-stuff/cool-spotify-blend/callback";
    const GRANT_TYPE: &str = "authorization_code";

    let (id, secret, redirect_uri) = read_from_env();

    let mut params = HashMap::new();
    params.insert("grant_type", GRANT_TYPE.to_owned());
    params.insert("code", code.to_owned());
    params.insert("redirect_uri", redirect_uri.to_owned());
    params.insert("client_id", id);
    params.insert("client_secret", secret);

    let client = reqwest::Client::new();
    let res = client
        .post(URL.to_owned())
        .query(&params)
        .header("Content-Type", "application/x-www-form-urlencoded")
        .header("Content-Length", 0)
        .send()
        .await?
        .json()
        .await?;

    Ok(res)
}

#[derive(Deserialize, Debug)]
struct RefreshTokenForIdRes {
    access_token: Token,
}

pub async fn refresh_token_for_id(spotify_id: &str, pool: &MySqlPool) -> Result<Token, JointErr> {
    // Get the refresh token from the database:
    let token = get_refresh_token_by_id(spotify_id, pool).await?;
    let (id, secret, _) = read_from_env();

    let mut params = HashMap::new();
    params.insert("grant_type", "refresh_token");
    params.insert("refresh_token", &token.0);
    params.insert("client_id", &id);
    params.insert("client_secret", &secret);

    let client = reqwest::Client::new();
    let res: RefreshTokenForIdRes = client
        .post(URL)
        .header("Content-Type", "application/x-www-form-urlencoded")
        .header("Content-Length", 0)
        .query(&params)
        .send()
        .await?
        .json()
        .await?;

    Ok(res.access_token)
}

#[derive(Debug, Deserialize)]
struct UserProfile {
    pub email: String,
    pub id: String,
}

async fn get_user_email(token: &Token) -> Result<UserProfile, reqwest::Error> {
    let client = reqwest::Client::new();
    let res = client
        .get(crate::BASE_URL.to_owned() + "/me")
        .header("Authorization", "Bearer ".to_owned() + &token.0)
        .send()
        .await?
        .json()
        .await?;

    Ok(res)
}

#[derive(Debug, Deserialize)]
#[non_exhaustive]
struct ClientCredentialsResponse {
    pub access_token: Token,
}

pub async fn get_client_credentials() -> Result<Token, reqwest::Error> {
    let client = reqwest::Client::new();
    let (id, secret, _) = read_from_env();

    let mut params = HashMap::new();
    params.insert("grant_type", "client_credentials".to_owned());

    let res: ClientCredentialsResponse = client
        .post("https://accounts.spotify.com/api/token")
        .query(&params)
        .header("Content-Type", "application/x-www-form-urlencoded")
        .header("Content-Length", 0)
        .basic_auth(&id, Some(&secret))
        .send()
        .await?
        .json()
        .await?;

    Ok(res.access_token)
}
